#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "TIFF.h"
#ifdef _WIN32
#include <TIFF/tiffio.h>
#else // !_WIN32
#include <tiffio.h>
#endif // _WIN32

void TIFFWriteColor( const char* fileName , ConstPointer( unsigned char ) pixels , int width , int height )
{
	TIFF* tiff = TIFFOpen( fileName , "w" );
	if( !tiff )																	fprintf( stderr , "TIFFOpen failed\n" )								,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_IMAGEWIDTH , width ) )					fprintf( stderr , "Can't set ImageWidth tag.\n" )					,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_IMAGELENGTH , height ) )					fprintf( stderr , "Can't set ImageLength tag.\n" )					,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_BITSPERSAMPLE , sizeof(char)*8 ) )		fprintf( stderr , "Can't set BitsPerSample tag.\n" )				,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_SAMPLESPERPIXEL , 3 ) )					fprintf( stderr , "Can't set SamplesPerPixel tag.\n" )				,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_COMPRESSION , COMPRESSION_LZW ) )			fprintf( stderr , "Can't set Compression tag.\n" )					,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_ROWSPERSTRIP , 1 ) )						fprintf( stderr , "Can't set RowsPerStrip tag.\n" )					,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_PLANARCONFIG , PLANARCONFIG_CONTIG ) )	fprintf( stderr , "Can't set PlanarConfiguration tag.\n" )			,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_PHOTOMETRIC , PHOTOMETRIC_RGB ) )			fprintf( stderr , "Can't set PhotometricInterpretation tag.\n" )	,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_RESOLUTIONUNIT , RESUNIT_NONE ) )			fprintf( stderr , "Can't set ResolutionUnit tag.\n" )				,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_ORIENTATION , ORIENTATION_TOPLEFT ) )		fprintf( stderr , "Can't set Orientation tag.\n" )					,	exit(0);

	Pointer( unsigned char ) _pixels = AllocPointer< unsigned char >( 3*width );
	for( int j=0 ; j<height ; j++ )
	{
		memcpy( _pixels , pixels + j * width * 3 , sizeof(unsigned char)*width*3 );
		TIFFWriteScanline( tiff , PointerAddress( _pixels ) , j , 0 );
	}
	FreePointer( _pixels );
	TIFFClose( tiff );
}
void TIFFWriteGray( const char* fileName , ConstPointer( unsigned char ) pixels , int width , int height )
{
	TIFF* tiff = TIFFOpen( fileName , "w" );
	if( !tiff )																	fprintf( stderr , "TIFFOpen failed\n" )								,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_IMAGEWIDTH , width ) )					fprintf( stderr , "Can't set ImageWidth tag.\n" )					,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_IMAGELENGTH , height ) )					fprintf( stderr , "Can't set ImageLength tag.\n" )					,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_BITSPERSAMPLE , sizeof(char)*8 ) )		fprintf( stderr , "Can't set BitsPerSample tag.\n" )				,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_SAMPLESPERPIXEL , 1 ) )					fprintf( stderr , "Can't set SamplesPerPixel tag.\n" )				,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_COMPRESSION , COMPRESSION_LZW ) )			fprintf( stderr , "Can't set Compression tag.\n" )					,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_ROWSPERSTRIP , 1 ) )						fprintf( stderr , "Can't set RowsPerStrip tag.\n" )					,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_PLANARCONFIG , PLANARCONFIG_CONTIG ) )	fprintf( stderr , "Can't set PlanarConfiguration tag.\n" )			,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_RESOLUTIONUNIT , RESUNIT_NONE ) )			fprintf( stderr , "Can't set ResolutionUnit tag.\n" )				,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_ORIENTATION , ORIENTATION_TOPLEFT ) )		fprintf( stderr , "Can't set Orientation tag.\n" )					,	exit(0);
	if( !TIFFSetField( tiff , TIFFTAG_PHOTOMETRIC , PHOTOMETRIC_MINISBLACK ) )	fprintf( stderr , "Can't set PhotometricInterpretation tag.\n" )	,	exit(0);

	Pointer( unsigned char ) _pixels = AllocPointer< unsigned char >( width );
	for( int j=0 ; j<height ; j++ )
	{
		memcpy( _pixels , pixels + j * width , sizeof(unsigned char)*width );
		TIFFWriteScanline( tiff , PointerAddress( _pixels ) , j , 0 );
	}
	FreePointer( _pixels );
	TIFFClose( tiff );
}

const double CharToInt16 = double((1<<16)-1) / double((1<<8)-1);
bool TIFFReadColor( const char* fileName , Pointer( unsigned char ) pixels , int width , int height )
{
	uint16 temp16;
	uint16 bitsPerSample , samplesPerPixel;
	TIFF* tiff = TIFFOpen( fileName , "r" );
	if( !tiff ) fprintf( stderr , "TIFFOpen failed\n" ) , exit( 0 );

	int _width , _height;
	TIFFGetField( tiff , TIFFTAG_IMAGEWIDTH  , &_width  );
	TIFFGetField( tiff , TIFFTAG_IMAGELENGTH , &_height );
	if( width!=_width || height!=_height ) fprintf( stderr , "[ERROR] TIFFReadColor: resolutions don't match: %d x %d != %d x %d\n" , width , height , _width , _height ) , exit( 0 );

	TIFFGetField( tiff , TIFFTAG_BITSPERSAMPLE , &bitsPerSample );
	if( bitsPerSample!=8 ) fprintf( stderr , "[ERROR] TIFFReadColor: Unsupported number of bits per sample: %d\n" , bitsPerSample ) , exit( 0 );
	TIFFGetField( tiff , TIFFTAG_SAMPLESPERPIXEL , &samplesPerPixel );
	if( samplesPerPixel!=1 && samplesPerPixel!=3 && samplesPerPixel!=4 ) fprintf( stderr , "[ERROR] TIFFReadColor: Unsupported number of samples per pixel: %d\n" , bitsPerSample ) , exit( 0 );
	int scanLineSize1 = TIFFScanlineSize( tiff ) , scanLineSize2 = ( bitsPerSample / 8 ) * samplesPerPixel * width;
	if( scanLineSize1!=scanLineSize2 ) fprintf( stderr , "[ERROR] TIFFReadColor: Scanline sizes do not agree: %d!=%d\n" , scanLineSize1 , scanLineSize2 ) , exit( 0 );
	TIFFGetField( tiff , TIFFTAG_PLANARCONFIG , &temp16 );
	if( temp16!=PLANARCONFIG_CONTIG ) fprintf( stderr , "[ERROR] TIFFReadColor: Planar configuration must be contiguous: %d != %d\n" , temp16 , PLANARCONFIG_CONTIG );
	tdata_t buff = _TIFFmalloc( scanLineSize1 );

	for( int j=0 ; j<height ; j++ )
	{
		TIFFReadScanline( tiff , buff , j );
		Pointer( unsigned char ) _pixels = pixels + j*width*3;

		Pointer( unsigned char ) buffer = GetPointer( (unsigned char*)buff , scanLineSize1 );
		if     ( samplesPerPixel==1 ) for( int i=0 ; i<width ; i++ ) _pixels[3*i] = _pixels[3*i+1] = _pixels[3*i+2] = buffer[i];
		else if( samplesPerPixel==3 ) for( int i=0 ; i<width ; i++ ) _pixels[3*i] = buffer[3*i] , _pixels[3*i+1] = buffer[3*i+1] , _pixels[3*i+2] = buffer[3*i+2];
		else if( samplesPerPixel==4 ) for( int i=0 ; i<width ; i++ ) _pixels[3*i] = buffer[4*i] , _pixels[3*i+1] = buffer[4*i+1] , _pixels[3*i+2] = buffer[4*i+2];
	}
	_TIFFfree( buff );
	TIFFClose( tiff );

	return true;
}
bool TIFFReadGray( const char* fileName , Pointer( unsigned char ) pixels , int width , int height )
{
	uint16 temp16;
	uint16 bitsPerSample , samplesPerPixel;
	TIFF* tiff = TIFFOpen( fileName , "r" );
	if( !tiff ) fprintf( stderr , "TIFFOpen failed\n" ) , exit( 0 );

	int _width , _height;
	TIFFGetField( tiff , TIFFTAG_IMAGEWIDTH  , &_width  );
	TIFFGetField( tiff , TIFFTAG_IMAGELENGTH , &_height );
	if( width!=_width || height!=_height ) fprintf( stderr , "[ERROR] TIFFReadColor: resolutions don't match: %d x %d != %d x %d\n" , width , height , _width , _height ) , exit( 0 );

	TIFFGetField( tiff , TIFFTAG_BITSPERSAMPLE , &bitsPerSample );
	if( bitsPerSample!=8 ) fprintf( stderr , "[ERROR] TIFFReadColor: Unsupported number of bits per sample: %d\n" , bitsPerSample ) , exit( 0 );
	TIFFGetField( tiff , TIFFTAG_SAMPLESPERPIXEL , &samplesPerPixel );
	if( samplesPerPixel!=1 && samplesPerPixel!=3 && samplesPerPixel!=4 ) fprintf( stderr , "[ERROR] TIFFReadColor: Unsupported number of samples per pixel: %d\n" , bitsPerSample ) , exit( 0 );
	int scanLineSize1 = TIFFScanlineSize( tiff ) , scanLineSize2 = ( bitsPerSample / 8 ) * samplesPerPixel * width;
	if( scanLineSize1!=scanLineSize2 ) fprintf( stderr , "[ERROR] TIFFReadColor: Scanline sizes do not agree: %d!=%d\n" , scanLineSize1 , scanLineSize2 ) , exit( 0 );
	TIFFGetField( tiff , TIFFTAG_PLANARCONFIG , &temp16 );
	if( temp16!=PLANARCONFIG_CONTIG ) fprintf( stderr , "[ERROR] TIFFReadColor: Planar configuration must be contiguous: %d != %d\n" , temp16 , PLANARCONFIG_CONTIG );
	tdata_t buff = _TIFFmalloc( scanLineSize1 );

	for( int j=0 ; j<height ; j++ )
	{
		TIFFReadScanline( tiff , buff , j );
		Pointer( unsigned char ) _pixels = pixels + j*width;

		Pointer( unsigned char ) buffer = GetPointer( (unsigned char*)buff , scanLineSize1 );
		if     ( samplesPerPixel==1 ) for( int i=0 ; i<width ; i++ ) _pixels[i] = buffer[i];
		else if( samplesPerPixel==3 ) for( int i=0 ; i<width ; i++ ) _pixels[i] = ( buffer[3*i] + buffer[3*i+1] + buffer[3*i+2] ) / 3;
		else if( samplesPerPixel==4 ) for( int i=0 ; i<width ; i++ ) _pixels[i] = ( buffer[4*i] + buffer[4*i+1] + buffer[4*i+2] ) / 3;
	}
	_TIFFfree( buff );
	TIFFClose( tiff );

	return true;
}

Pointer( unsigned char ) TIFFReadColor( const char* fileName , int& width , int& height )
{
	uint16 temp16;
	uint16 bitsPerSample , samplesPerPixel;
	TIFF* tiff = TIFFOpen( fileName , "r" );
	if( !tiff ) fprintf( stderr , "TIFFOpen failed\n" ) , exit( 0 );

	TIFFGetField( tiff , TIFFTAG_IMAGEWIDTH  , &width  );
	TIFFGetField( tiff , TIFFTAG_IMAGELENGTH , &height );
	Pointer( unsigned char ) pixels = AllocPointer< unsigned char >( width * height * 3 );
	if( !pixels ) fprintf( stderr , "[ERROR] Failed to allocate pixels: %d X %d\n" , width , height ) , exit( 0 );

	TIFFGetField( tiff , TIFFTAG_BITSPERSAMPLE , &bitsPerSample );
	if( bitsPerSample!=8 ) fprintf( stderr , "[ERROR] TIFFReadColor: Unsupported number of bits per sample: %d\n" , bitsPerSample ) , exit( 0 );
	TIFFGetField( tiff , TIFFTAG_SAMPLESPERPIXEL , &samplesPerPixel );
	if( samplesPerPixel!=1 && samplesPerPixel!=3 && samplesPerPixel!=4 ) fprintf( stderr , "[ERROR] TIFFReadColor: Unsupported number of samples per pixel: %d\n" , bitsPerSample ) , exit( 0 );
	int scanLineSize1 = TIFFScanlineSize( tiff ) , scanLineSize2 = ( bitsPerSample / 8 ) * samplesPerPixel * width;
	if( scanLineSize1!=scanLineSize2 ) fprintf( stderr , "[ERROR] TIFFReadColor: Scanline sizes do not agree: %d!=%d\n" , scanLineSize1 , scanLineSize2 ) , exit( 0 );
	TIFFGetField( tiff , TIFFTAG_PLANARCONFIG , &temp16 );
	if( temp16!=PLANARCONFIG_CONTIG ) fprintf( stderr , "[ERROR] TIFFReadColor: Planar configuration must be contiguous: %d != %d\n" , temp16 , PLANARCONFIG_CONTIG );
	tdata_t buff = _TIFFmalloc( scanLineSize1 );

	for( int j=0 ; j<height ; j++ )
	{
		TIFFReadScanline( tiff , buff , j );
		Pointer( unsigned char ) _pixels = pixels + j*width*3;

		Pointer( unsigned char ) buffer = GetPointer( (unsigned char*)buff , scanLineSize1 );
		if     ( samplesPerPixel==1 ) for( int i=0 ; i<width ; i++ ) _pixels[3*i] = _pixels[3*i+1] = _pixels[3*i+2] = buffer[i];
		else if( samplesPerPixel==3 ) for( int i=0 ; i<width ; i++ ) _pixels[3*i] = buffer[3*i] , _pixels[3*i+1] = buffer[3*i+1] , _pixels[3*i+2] = buffer[3*i+2];
		else if( samplesPerPixel==4 ) for( int i=0 ; i<width ; i++ ) _pixels[3*i] = buffer[4*i] , _pixels[3*i+1] = buffer[4*i+1] , _pixels[3*i+2] = buffer[4*i+2];
	}
	_TIFFfree( buff );
	TIFFClose( tiff );

	return pixels;
}
Pointer( unsigned char ) TIFFReadGray( const char* fileName , int& width , int& height )
{
	uint16 temp16;
	uint16 bitsPerSample , samplesPerPixel;
	TIFF* tiff = TIFFOpen( fileName , "r" );
	if( !tiff ) fprintf( stderr , "TIFFOpen failed\n" ) , exit( 0 );

	TIFFGetField( tiff , TIFFTAG_IMAGEWIDTH  , &width  );
	TIFFGetField( tiff , TIFFTAG_IMAGELENGTH , &height );
	Pointer( unsigned char ) pixels = AllocPointer< unsigned char >( width * height * 3 );
	if( !pixels ) fprintf( stderr , "[ERROR] Failed to allocate pixels: %d X %d\n" , width , height ) , exit( 0 );

	TIFFGetField( tiff , TIFFTAG_BITSPERSAMPLE , &bitsPerSample );
	if( bitsPerSample!=8 ) fprintf( stderr , "[ERROR] TIFFReadColor: Unsupported number of bits per sample: %d\n" , bitsPerSample ) , exit( 0 );
	TIFFGetField( tiff , TIFFTAG_SAMPLESPERPIXEL , &samplesPerPixel );
	if( samplesPerPixel!=1 && samplesPerPixel!=3 && samplesPerPixel!=4 ) fprintf( stderr , "[ERROR] TIFFReadColor: Unsupported number of samples per pixel: %d\n" , bitsPerSample ) , exit( 0 );
	int scanLineSize1 = TIFFScanlineSize( tiff ) , scanLineSize2 = ( bitsPerSample / 8 ) * samplesPerPixel * width;
	if( scanLineSize1!=scanLineSize2 ) fprintf( stderr , "[ERROR] TIFFReadColor: Scanline sizes do not agree: %d!=%d\n" , scanLineSize1 , scanLineSize2 ) , exit( 0 );
	TIFFGetField( tiff , TIFFTAG_PLANARCONFIG , &temp16 );
	if( temp16!=PLANARCONFIG_CONTIG ) fprintf( stderr , "[ERROR] TIFFReadColor: Planar configuration must be contiguous: %d != %d\n" , temp16 , PLANARCONFIG_CONTIG );
	tdata_t buff = _TIFFmalloc( scanLineSize1 );

	for( int j=0 ; j<height ; j++ )
	{
		TIFFReadScanline( tiff , buff , j );
		Pointer( unsigned char ) _pixels = pixels + j*width;

		Pointer( unsigned char ) buffer = GetPointer( (unsigned char*)buff , scanLineSize1 );
		if     ( samplesPerPixel==1 ) for( int i=0 ; i<width ; i++ ) _pixels[i] = _pixels[3*i+1] = _pixels[3*i+2] = buffer[i];
		else if( samplesPerPixel==3 ) for( int i=0 ; i<width ; i++ ) _pixels[i] = ( buffer[3*i] + buffer[3*i+1] + buffer[3*i+2] ) / 3;
		else if( samplesPerPixel==4 ) for( int i=0 ; i<width ; i++ ) _pixels[i] = ( buffer[4*i] + buffer[4*i+1] + buffer[4*i+2] ) / 3;
	}
	_TIFFfree( buff );
	TIFFClose( tiff );
	return pixels;
}
bool TIFFGetImageSize( const char* fileName , int& width , int& height , int& channelsPerPixel , int& bitsPerChannel )
{
	uint16 bitsPerSample , samplesPerPixel;
	TIFF* tiff = TIFFOpen( fileName , "r" );
	if( !tiff ) fprintf( stderr , "TIFFOpen failed\n" ) , exit( 0 );
	TIFFGetField( tiff , TIFFTAG_IMAGEWIDTH  , &width  );
	TIFFGetField( tiff , TIFFTAG_IMAGELENGTH , &height );
	TIFFGetField( tiff , TIFFTAG_BITSPERSAMPLE , &bitsPerSample );
	TIFFGetField( tiff , TIFFTAG_SAMPLESPERPIXEL , &samplesPerPixel );
	TIFFClose( tiff );
	channelsPerPixel = samplesPerPixel;
	bitsPerChannel = (int)bitsPerSample;

	return true;
}
bool TIFFGetImageSize( const char* fileName , int& width , int& height ){ int bitsPerChannel , channelsPerPixel ; return TIFFGetImageSize( fileName , width , height , channelsPerPixel , bitsPerChannel ); }
